// LEDマトリクス64x32(RGB) for PIC32MX250F128B 12ビット色対応 Rev.2
// PMPおよびDMA機能使用
// 横64ドット分×4フレーム×16行分のDMA用バッファを確保(フレームに重み付けあり)
// ピン接続
// PMD0-PMD5 → B1,G1,R1,B2,G2,R2 (PMD6,PMD7は不使用)
// PMWR → CLK
// PMRD 不使用
// PMAx 不使用
// A:RB10 B:RB11 C:RB13 D:RB14
// LAT:RA2 OE:RB4

#include <xc.h>
#include <sys/attribs.h>
#include "led64x32-3.h"

#define PLANENUM 4
#define PLANEORDERNUM 15
unsigned char LEDBUF[64*16*PLANENUM] __attribute__ ((aligned (4)));
//LEDBUF[] LEDの6本のシフトレジスタへ連続転送するそのままの形式でデータ保有

unsigned short VRAM[64*32] __attribute__ ((aligned (4)));
unsigned short TEMP[64] __attribute__ ((aligned (4)));
volatile int T2IRcounter,drawcount;
volatile unsigned char *LEDBUFp;
volatile int LEDsection,planecount;
const unsigned char planeorder[]={3,2,3,1,3,2,3,0,3,2,3,1,3,2,3};

// タイマー2割り込み処理ルーチン
void __ISR(_TIMER_2_VECTOR, IPL4AUTO) T2Interrupt(void) {
	DCH1SSA = (unsigned int)LEDBUFp & 0x1FFFFFFF; //転送元にLEDバッファを指定
	DCH1CON = 0x0080; // チャネル1有効
	DCH1ECONSET = 0x80; // DMA1転送開始
	T2IRcounter++;
	IFS0bits.T2IF = 0; // Reset interrupt flag
}

// DMA1割り込み
// 転送元サイズの転送完了時に割り込み発生
void __ISR(_DMA1_VECTOR, IPL4AUTO) _DMA1Interrupt ()
{
//	LATBSET=0x0010;//OE=ON(display OFF)
	LATB=((LEDsection & 0x0c)<<11)+((LEDsection & 0x03)<<10)+0x10; //set address & OE=ON
	LATASET=0x0004;//LAT=ON
	asm("nop");
	LATACLR=0x0004;//LAT=OFF (latch data)
	asm("nop");
	LATBCLR=0x0010;//OE=OFF(display ON)
	planecount++;
	if(planecount==PLANEORDERNUM){
		planecount=0;
		LEDsection++;
		if(LEDsection==16){
			LEDsection=0;
			drawcount++;
		}
	}
	LEDBUFp=LEDBUF+planeorder[planecount]*64*16+LEDsection*64;
	DCH1INTCLR=0x0080;
	IFS1bits.DMA1IF=0; // Clear DMA ch1 IF
}
void LEDBUFclear(void){
	unsigned int *p;
	for(p=(unsigned int *)LEDBUF;p<(unsigned int *)(LEDBUF+64*16*PLANENUM);p++) *p=0;
}
void LEDBUFsetp(int x,int y,int c){
	//LEDバッファ上に点を描画
	// c:RRRRGGGGBBBB
	unsigned char *p;
	int i;
	if(y<16){
		p=LEDBUF+x+y*64;
		for(i=0;i<PLANENUM;i++){
			*p=(*p&0x38)|((c&1)+((c&0x10)>>3)+((c&0x100)>>6));
			c>>=1;
			p+=64*16;
		}
	}
	else{
		p=LEDBUF+x+(y-16)*64;
		for(i=0;i<PLANENUM;i++){
			*p=(*p&0x7)|(((c&1)<<3)+(c&0x10)+((c&0x100)>>3));
			c>>=1;
			p+=64*16;
		}
	}
}
void xferVRAM2VBUF(void){
	//VRAMからLEDBUFへデータ形式変換し転送
	int x,y,c;
	unsigned short *p;
	p=VRAM;
	for(y=0;y<32;y++){
		for(x=0;x<64;x++){
			LEDBUFsetp(x,y,*p++);
		}
	}
}
void cls(void){
	unsigned int *p;
	for(p=(unsigned int *)VRAM;p<(unsigned int *)(VRAM+64*32);p++) *p=0;
}
void pset(int x,int y,int c){
	// c:RRRRGGGGBBBB
	if((unsigned int)x<64 && (unsigned int)y<32) VRAM[y*64+x]=c;
}
void vscroll(void){
	unsigned int *sp,*dp;
	int i;
	sp=(unsigned int *)VRAM;
	dp=(unsigned int *)TEMP;
	for(i=0;i<32;i++) *dp++=*sp++;
	dp=(unsigned int *)VRAM;
	for(i=0;i<32*31;i++) *dp++=*sp++;
	sp=(unsigned int *)TEMP;
	for(i=0;i<32;i++) *dp++=*sp++;
}
void putfont(int x,int y,int c,unsigned char ch){
	unsigned int f;
	int i,j;
	for(i=0;i<8;i++){
		if(i+y<0) continue;
		if(i+y>=32) break;
		f=FontData[ch*8+i];
		for(j=0;j<8;j++){
			f<<=1;
			if(j+x<0) continue;
			if(j+x>=64) break;
			if(f & 0x100) VRAM[(i+y)*64+j+x]=c;
			else VRAM[(i+y)*64+j+x]=0;
		}
	}
}
void init_led64x32(void) {
	//タイマー2割り込み設定
	T2CON = 0x00; // prescale 1:1, 16bit mode
	PR2 = 3333; // 0.069ms,14.4KHz n/48MHz(s) 60frame/s
	TMR2 = 0; // clear Timer 2 counter

	//PMP設定
	PMCON = 0x0202; // PMWR有効、PMRD無効、アクティブH
	PMMODE = 0x22d0; // PM割り込み有効、8bit マスターモード2,WAITB=4,WAITM=5,WAITE=0
	PMADDR = 0x0000; // CSx無効
	PMAEN = 0x0000; // PMAxは全て無効
	PMCONSET = 0x8000; // PMP有効化

	DMACON = 0x8000;

	//DMAチャネル1設定
	DCH1CON = 0x0000; //チャネル1無効
	DCH1ECON = (_PMP_IRQ << 8) | 0x10; //PMP割り込みでトリガ
	DCH1DSA = (unsigned int)&PMDIN & 0x1FFFFFFF; //転送先をPMP出力に設定
	DCH1SSIZ = 64; //転送元サイズ
	DCH1DSIZ = 1; //転送先サイズ
	DCH1CSIZ = 1; //転送セルサイズ
	DCH1INT = 0x00800000; //ブロック転送終了割り込み

	IPC2bits.T2IP = 4; // T2割り込み優先度4に設定
	IFS0bits.T2IF = 0; // T2割り込みフラグをクリア
	IEC0bits.T2IE = 1; // T2割り込み有効化
	IPC10bits.DMA1IP = 4; // DMA1割り込み優先度4に設定
	IFS1bits.DMA1IF = 0; // DMA1割り込みフラグをクリア
	IEC1bits.DMA1IE = 1; // DMA1割り込み有効化

	LEDBUFclear();
	cls();
	T2IRcounter=0;
	drawcount=0;
	planecount=0;
	LEDBUFp=LEDBUF+planeorder[planecount]*64*16;
	LEDsection=0;

	T2CONbits.ON = 1; //タイマー2開始
}
